#!/bin/bash

#########
# This script checks all files from packages installed through RPM.
# It can also recognize files that do not belong to any installed package.
#
# Author: Diego S. Gomes
# License: GPL
#########

#if [ $# -lt 1 -o $# -gt 2 ]; then
if [ "$#" -eq 0 ]; then
    echo "Usage: $0 [-cron] distpath [rpmroot]"
    echo "   - distpath: path to the directory containing files/directories"\
         "to be checked."
    echo "   - rpmroot: optionally says where RPM dbs are located (will use"\
         "RPM_ROOT env variable by default)"
    exit 1
fi

if [ "$1" = "-cron" ]; then
    ISCRONJOB=true
    shift 1  # $1 now refers to the next argument
fi

if [ "${1:0:1}" = "/" ]; then
	DISTPATH=${1};
else
	DISTPATH=$(pwd)/${1};
fi

if [ -n "${2}" ]; then
	RPM_ROOT=${2};
fi

if [ -z "${RPM_ROOT}" -o ! -d "${RPM_ROOT}" ]; then
    echo "Could not locate RPM database. Check if RPM_ROOT is defined correctly."
    exit 2
fi


## For each file under the distpath:
## - It does not belong to any distribution package; or
## - It belongs to a package and is correct; or
## - It belongs to a package, but was modified;
## Also, package files that should be under distpath may be missing.

# Checks distro packages that contain files under distpath.
# Directories are only taken into account if they are empty.
FILE_TMPLIST=/tmp/chkrpm$$-tmpfile
find ${DISTPATH} | 
(read PREVLINE; 
while read LINE; do 
	if [[ ! "$LINE" =~ "$PREVLINE" ]]; then 
		rpm -qf $PREVLINE; 
	fi; 
	PREVLINE=$LINE; 
done; 
rpm -qf $PREVLINE) | sort -u > ${FILE_TMPLIST}

#FILE_TMPLIST=/tmp/chkrpm$$-tmpfile
#for FILE in $(find ${DISTPATH}); do
#	rpm -qf $FILE;
#done | sort -u > ${FILE_TMPLIST}

# Gets the list of distro packages containing files under distpath
FILE_PKGLIST=/tmp/chkrpm$$-pkglist
grep -v "file " ${FILE_TMPLIST} > ${FILE_PKGLIST}

# Gets the list of files that do not belong to any package
FILE_ORPHANLIST=/tmp/chkrpm$$-orphanlist
grep "file " ${FILE_TMPLIST} | cut -d" " -f2 > ${FILE_ORPHANLIST}

# Checks files in each distro package
FILE_VERIFYOUT=/tmp/chkrpm$$-verifyout
for PKG in $(cat ${FILE_PKGLIST}); do
	rpm -V --nodeps --noscript ${PKG};
done > ${FILE_VERIFYOUT} 

# From the distro files checked, it gets only those under DISTPATH
FILE_VALIDLIST=/tmp/chkrpm$$-validlist 
while read LINE; do
	POSPATH=$(expr index "$LINE" "/");
	PATHSUBSTR=$(expr substr "$LINE" $POSPATH ${#DISTPATH})
	if [ "${PATHSUBSTR}" = "${DISTPATH}" ]; then
		# The line contains a file under DISTPATH
		echo ${LINE};
	fi
done < ${FILE_VERIFYOUT} > ${FILE_VALIDLIST}

## Now process the list of checked files under DISTPATH
#

# Gets the list of missing files
FILE_MISSLIST=/tmp/chkrpm$$-misslist
grep "missing" ${FILE_VALIDLIST} > ${FILE_MISSLIST}

# Interprets the rpm output for the remaining files
FILE_ATTRSIZE=/tmp/chkrpm$$-attrsize
FILE_ATTRMODE=/tmp/chkrpm$$-attrmode
FILE_ATTRMD5=/tmp/chkrpm$$-attrmd5
FILE_ATTRMAJORMINOR=/tmp/chkrpm$$-attrmajorminor
FILE_ATTRLINK=/tmp/chkrpm$$-attrlink
FILE_ATTROWNER=/tmp/chkrpm$$-attrowner
FILE_ATTRGROUP=/tmp/chkrpm$$-attrgroup
FILE_ATTRTIME=/tmp/chkrpm$$-attrtime
FILE_MODIFIED=/tmp/chkrpm$$-modified
FILE_ATTRS=/tmp/chkrpm$$-attrs
grep -v "missing" ${FILE_VALIDLIST} | while read LINE; do
	POS=$(expr index "$LINE" "/");
	FILE=${LINE:$POS-1};

	# Checks if file size changed
	#if [ "${LINE:0:1}" = "S" ]; then
	#	echo ${FILE} >> ${FILE_ATTRSIZE};
	#fi 

	# Checks if file mode changed
        #if [ "${LINE:1:1}" = "M" ]; then
        #        echo ${FILE} >> ${FILE_ATTRMODE};
        #fi

        # Checks if file md5 checksum matches
        #if [ "${LINE:2:1}" = "5" ]; then
        #        echo ${FILE} >> ${FILE_ATTRMD5};
        #fi

        # Checks if file major/minor number changed
        #if [ "${LINE:3:1}" = "D" ]; then
        #        echo ${FILE} >> ${FILE_ATTRMAJORMINOR};
        #fi

        # Checks if file symbolic link content changed
        #if [ "${LINE:4:1}" = "L" ]; then
        #        echo ${FILE} >> ${FILE_ATTRLINK};
        #fi

        # Checks if file owner changed
        #if [ "${LINE:5:1}" = "U" ]; then
        #        #echo ${FILE} >> ${FILE_ATTROWNER};
        #fi

        # Checks if file group changed
        #if [ "${LINE:6:1}" = "G" ]; then
        #        echo ${FILE} >> ${FILE_ATTRGROUP};
        #fi

        # Checks if file modification time changed
        #if [ "${LINE:7:1}" = "T" ]; then
        #        echo ${FILE} >> ${FILE_ATTRTIME};
        #fi

        # Checks if file changed (size, md5, link content, modification time, mode)
        if [ "${LINE:0:1}" = "S" -o "${LINE:2:1}" = "5" -o "${LINE:4:1}" = "L" -o "${LINE:7:1}" = "T" -o "${LINE:1:1}" = "M" ]; then
                echo ${FILE} >> ${FILE_MODIFIED};
        fi

	# Checks if file mode/ownership changed (owner, group, mode, major/minor)
        #if [ "${LINE:5:1}" = "U" -o "${LINE:6:1}" = "G" -o "${LINE:1:1}" = "M" -o "${LINE:3:1}" = "D" ]; then
        #        echo ${FILE} >> ${FILE_ATTRS};
        #fi

done

## Builds and prints a report
#
if [ ! "$ISCRONJOB" -o -s ${FILE_ORPHANLIST} -o -s ${FILE_MISSLIST} -o -s ${FILE_MODIFIED} ]; then
	echo "------[Analysis for files under ${DISTPATH}]"
	TOTAL=$(find ${DISTPATH}|wc -l)
	echo "${TOTAL} file(s) found"
	ORPHANS=$(cat ${FILE_ORPHANLIST}|wc -l)
	echo "${ORPHANS} unknown"
	PACKAGES=$(cat ${FILE_PKGLIST}|wc -l)
	let CONTAINED=TOTAL-ORPHANS
	#echo "${CONTAINED} are contained in ${PACKAGES} packages"
	#MODIFIED=$(grep -v "missing" ${FILE_VALIDLIST}|wc -l)
        [ -f ${FILE_MODIFIED} ] && MODIFIED=$(cat ${FILE_MODIFIED}|wc -l) || MODIFIED="0"
	echo "${MODIFIED} modified"
	MISSING=$(cat ${FILE_MISSLIST}|wc -l)
	echo "${MISSING} missing"
	#echo "------[end of statistics]"

	[ -s ${FILE_ORPHANLIST} ] && echo -e "\n------[Unknown files]" && cat ${FILE_ORPHANLIST} 
	[ -s ${FILE_MISSLIST} ]   && echo -e "\n------[Missing files]"                           && cat ${FILE_MISSLIST} 
	[ -s ${FILE_MODIFIED} ]   && echo -e "\n------[Modified files (size, md5, modification time or mode)]" && cat ${FILE_MODIFIED}  
	[ -s ${FILE_ATTRS} -a ! "$ISCRONJOB" ]   && echo -e "\n------[List of files with ownership or mode changed]" && cat ${FILE_ATTRS} 
	[ -s ${FILE_PKGLIST} ]    && echo -e "\n------[Analysed packages]"  && cat ${FILE_PKGLIST} 
fi

## Cleans some stuff 
#
rm -f ${FILE_PKGLIST}
rm -f ${FILE_ORPHANLIST}
rm -f ${FILE_MISSLIST}
rm -f ${FILE_MODIFIED}
rm -f ${FILE_ATTRS}
#rm -f /tmp/chkrpm*

exit 0

